import torch
import torch.nn as nn
import numpy as np
#from numba import jit

class HDSNRandomLoss(nn.Module):
    """
    Hieratical depth spatial normalization loss.
    Replace the original grid masks with the random created masks.
    loss = MAE((d-median(d)/s - (d'-median(d'))/s'), s = mean(d- median(d))
    """
    def __init__(self, loss_weight=1.0, random_num=32, data_type=['sfm', 'stereo', 'lidar', 'denselidar', 'denselidar_nometric','denselidar_syn'], disable_dataset=['MapillaryPSD'], sky_id=142, batch_limit=8, **kwargs):
        super(HDSNRandomLoss, self).__init__()
        self.loss_weight = loss_weight
        self.random_num = random_num
        self.data_type = data_type
        self.sky_id = sky_id
        self.batch_limit = batch_limit
        self.eps = 1e-6
        self.disable_dataset = disable_dataset

    def get_random_masks_for_batch(self, image_size: list)-> torch.Tensor:
        height, width = image_size
        crop_h_min = int(0.125 * height)
        crop_h_max = int(0.5 * height)
        crop_w_min = int(0.125 * width)
        crop_w_max = int(0.5 * width)
        h_max = height - crop_h_min
        w_max = width - crop_w_min
        crop_height = np.random.choice(np.arange(crop_h_min, crop_h_max), self.random_num, replace=False)
        crop_width = np.random.choice(np.arange(crop_w_min, crop_w_max), self.random_num, replace=False)
        crop_y = np.random.choice(h_max, self.random_num, replace=False)
        crop_x = np.random.choice(w_max, self.random_num, replace=False)
        crop_y_end = crop_height + crop_y
        crop_y_end[crop_y_end>=height] = height
        crop_x_end = crop_width + crop_x
        crop_x_end[crop_x_end>=width] = width

        mask_new = torch.zeros((self.random_num,  height, width), dtype=torch.bool, device="cuda") #.cuda() #[N, H, W]
        for i in range(self.random_num):
           mask_new[i, crop_y[i]:crop_y_end[i], crop_x[i]:crop_x_end[i]] = True

        return mask_new
        #return crop_y, crop_y_end, crop_x, crop_x_end
    
    def reorder_sem_masks(self, sem_label):
        # reorder the semantic mask of a batch
        assert sem_label.ndim == 3
        semantic_ids = torch.unique(sem_label[(sem_label>0) & (sem_label != self.sky_id)])
        sem_masks = [sem_label == id for id in semantic_ids]
        if len(sem_masks) == 0:
            # no valid semantic labels
            out = sem_label > 0
            return out

        sem_masks = torch.cat(sem_masks, dim=0)
        mask_batch = torch.sum(sem_masks.reshape(sem_masks.shape[0], -1), dim=1) > 500
        sem_masks = sem_masks[mask_batch]
        if sem_masks.shape[0] > self.random_num:
            balance_samples = np.random.choice(sem_masks.shape[0], self.random_num, replace=False)
            sem_masks = sem_masks[balance_samples, ...]
        
        if sem_masks.shape[0] == 0:
            # no valid semantic labels
            out = sem_label > 0
            return out

        if sem_masks.ndim == 2:
            sem_masks = sem_masks[None, :, :]
        return sem_masks
  
    def ssi_mae(self, prediction, target, mask_valid):
        B, C, H, W = target.shape
        prediction_nan = prediction.clone().detach()
        target_nan = target.clone()
        prediction_nan[~mask_valid] = float('nan')
        target_nan[~mask_valid] = float('nan')

        valid_pixs = mask_valid.reshape((B, C,-1)).sum(dim=2, keepdims=True) + 1e-10
        valid_pixs = valid_pixs[:, :, :, None]

        gt_median = target_nan.reshape((B, C,-1)).nanmedian(2, keepdims=True)[0].unsqueeze(-1) # [b,c,h,w]
        gt_median[torch.isnan(gt_median)] = 0
        gt_diff = (torch.abs(target - gt_median) ).reshape((B, C, -1))
        gt_s = gt_diff.sum(dim=2)[:, :, None, None] / valid_pixs
        gt_trans = (target - gt_median) / (gt_s + self.eps)

        pred_median = prediction_nan.reshape((B, C,-1)).nanmedian(2, keepdims=True)[0].unsqueeze(-1) # [b,c,h,w]
        pred_median[torch.isnan(pred_median)] = 0
        pred_diff = (torch.abs(prediction - pred_median)).reshape((B, C, -1))
        pred_s = pred_diff.sum(dim=2)[:, :, None, None] / valid_pixs
        pred_trans = (prediction - pred_median) / (pred_s + self.eps)

        loss_sum = torch.sum(torch.abs(gt_trans - pred_trans)*mask_valid)
        return loss_sum
    
    def conditional_ssi_mae(self, prediction, target, mask_valid):
        B, C, H, W = target.shape
        conditional_rank_ids = np.random.choice(B, B, replace=False)

        prediction_nan = prediction.clone()
        target_nan = target.clone()
        prediction_nan[~mask_valid] = float('nan')
        target_nan[~mask_valid] = float('nan')

        valid_pixs = mask_valid.reshape((B, C,-1)).sum(dim=2, keepdims=True) + self.eps
        valid_pixs = valid_pixs[:, :, :, None].contiguous()

        gt_median = target_nan.reshape((B, C,-1)).nanmedian(2, keepdims=True)[0].unsqueeze(-1) # [b,c,h,w]
        gt_median[torch.isnan(gt_median)] = 0
        gt_diff = (torch.abs(target - gt_median) * mask_valid).reshape((B, C,-1))
        gt_s = gt_diff.sum(dim=2)[:, :, None, None].contiguous() / valid_pixs

        # in case some batches have no valid pixels
        gt_s_small_mask = gt_s < (torch.mean(gt_s)*0.1)
        gt_s[gt_s_small_mask] = torch.mean(gt_s)
        gt_trans = (target - gt_median[conditional_rank_ids]) / (gt_s[conditional_rank_ids] + self.eps)

        pred_median = prediction_nan.reshape((B, C,-1)).nanmedian(2, keepdims=True)[0].unsqueeze(-1) # [b,c,h,w]
        pred_median[torch.isnan(pred_median)] = 0
        pred_diff = (torch.abs(prediction - pred_median) * mask_valid).reshape((B, C,-1))
        pred_s = pred_diff.sum(dim=2)[:, :, None, None].contiguous() / valid_pixs
        pred_s[gt_s_small_mask] = torch.mean(pred_s)
        pred_trans = (prediction - pred_median[conditional_rank_ids]) / (pred_s[conditional_rank_ids] + self.eps)

        loss_sum = torch.sum(torch.abs(gt_trans - pred_trans)*mask_valid)
        # print(torch.abs(gt_trans - pred_trans)[mask_valid])
        return loss_sum


    def forward(self, prediction, target, mask=None, sem_mask=None, **kwargs):
        """
        Calculate loss.
        """
        B, C, H, W = target.shape
        
        loss = 0.0
        valid_pix = 0.0

        device = target.device
        
        batches_dataset = kwargs['dataset']
        self.batch_valid = torch.tensor([1 if batch_dataset not in self.disable_dataset else 0 \
            for batch_dataset in batches_dataset], device=device)[:,None,None,None]

        batch_limit = self.batch_limit
        
        random_sample_masks = self.get_random_masks_for_batch((H, W)) # [N, H, W]
        for i in range(B):
            # each batch
            mask_i = mask[i, ...] #[1, H, W]
            if self.batch_valid[i, ...] < 0.5:
                loss += 0 * torch.sum(prediction[i, ...])
                valid_pix += 0 * torch.sum(mask_i)
                continue

            pred_i = prediction[i, ...].unsqueeze(0).repeat(batch_limit, 1, 1, 1)
            target_i = target[i, ...].unsqueeze(0).repeat(batch_limit, 1, 1, 1)

            # get semantic masks
            sem_label_i = sem_mask[i, ...] if sem_mask is not None else None
            if sem_label_i is not None:
                sem_masks = self.reorder_sem_masks(sem_label_i) # [N, H, W]
                random_sem_masks = torch.cat([random_sample_masks, sem_masks], dim=0)
            else:
                random_sem_masks = random_sample_masks
            #random_sem_masks = random_sample_masks


            sampled_masks_num = random_sem_masks.shape[0]
            loops = int(np.ceil(sampled_masks_num / batch_limit))
            conditional_rank_ids = np.random.choice(sampled_masks_num, sampled_masks_num, replace=False)

            for j in range(loops):
                mask_random_sem_loopi = random_sem_masks[j*batch_limit:(j+1)*batch_limit, ...]
                mask_sample = (mask_i & mask_random_sem_loopi).unsqueeze(1) # [N, 1, H, W]
                loss += self.ssi_mae(
                    prediction=pred_i[:mask_sample.shape[0], ...], 
                    target=target_i[:mask_sample.shape[0], ...], 
                    mask_valid=mask_sample)
                valid_pix += torch.sum(mask_sample)

                # conditional ssi loss
                # rerank_mask_random_sem_loopi = random_sem_masks[conditional_rank_ids, ...][j*batch_limit:(j+1)*batch_limit, ...]
                # rerank_mask_sample = (mask_i & rerank_mask_random_sem_loopi).unsqueeze(1) # [N, 1, H, W]
                # loss_cond = self.conditional_ssi_mae(
                #     prediction=pred_i[:rerank_mask_sample.shape[0], ...], 
                #     target=target_i[:rerank_mask_sample.shape[0], ...], 
                #     mask_valid=rerank_mask_sample)
                # print(loss_cond / (torch.sum(rerank_mask_sample) + 1e-10), loss_cond, torch.sum(rerank_mask_sample))
                # loss += loss_cond
                # valid_pix += torch.sum(rerank_mask_sample)

        # crop_y, crop_y_end, crop_x, crop_x_end = self.get_random_masks_for_batch((H, W)) # [N,]
        # for j in range(B):
        #     for i in range(self.random_num):
        #         mask_crop = mask[j, :, crop_y[i]:crop_y_end[i], crop_x[i]:crop_x_end[i]][None, ...] #[1, 1, crop_h, crop_w]
        #         target_crop = target[j, :, crop_y[i]:crop_y_end[i], crop_x[i]:crop_x_end[i]][None, ...]
        #         pred_crop = prediction[j, :, crop_y[i]:crop_y_end[i], crop_x[i]:crop_x_end[i]][None, ...]
        #         loss += self.ssi_mae(prediction=pred_crop, target=target_crop, mask_valid=mask_crop)
        #         valid_pix += torch.sum(mask_crop)
        
        # the whole image
        mask = mask * self.batch_valid.bool()
        loss += self.ssi_mae(
                    prediction=prediction, 
                    target=target, 
                    mask_valid=mask)
        valid_pix += torch.sum(mask)
        loss = loss / (valid_pix + self.eps)
        if torch.isnan(loss).item() | torch.isinf(loss).item():
            loss = 0 * torch.sum(prediction)
            print(f'HDSNL NAN error, {loss}, valid pix: {valid_pix}')
        return loss * self.loss_weight
    
if __name__ == '__main__':
    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    ssil = HDSNRandomLoss()
    pred = torch.rand((8, 1, 256, 512)).cuda()
    gt = torch.rand((8, 1, 256, 512)).cuda()#torch.zeros_like(pred).cuda() #
    gt[1:, :, 100:256, 100:350] = -1
    gt[:2, ...] = -1
    mask = gt > 0
    sem_mask = np.random.randint(-1, 200, (8, 1, 256, 512))
    sem_mask[sem_mask>0] = -1
    sem_mask_torch = torch.from_numpy(sem_mask).cuda()

    out = ssil(pred, gt, mask, sem_mask_torch)
    print(out)



# MIT License

# Copyright (c) 2022 Intelligent Systems Lab Org

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# File author: Shariq Farooq Bhat

import itertools

import torch
import torch.nn as nn
from zoedepth.models.depth_model import DepthModel
from zoedepth.models.base_models.midas import MidasCore
from zoedepth.models.layers.attractor import AttractorLayer, AttractorLayerUnnormed
from zoedepth.models.layers.dist_layers import ConditionalLogBinomial
from zoedepth.models.layers.localbins_layers import (Projector, SeedBinRegressor,
                                            SeedBinRegressorUnnormed)
from zoedepth.models.model_io import load_state_from_resource


class ZoeDepth(DepthModel):
    def __init__(self, core,  n_bins=64, bin_centers_type="softplus", bin_embedding_dim=128, min_depth=1e-3, max_depth=10,
                 n_attractors=[16, 8, 4, 1], attractor_alpha=300, attractor_gamma=2, attractor_kind='sum', attractor_type='exp', min_temp=5, max_temp=50, train_midas=True,
                 midas_lr_factor=10, encoder_lr_factor=10, pos_enc_lr_factor=10, inverse_midas=False, **kwargs):
        """ZoeDepth model. This is the version of ZoeDepth that has a single metric head

        Args:
            core (models.base_models.midas.MidasCore): The base midas model that is used for extraction of "relative" features
            n_bins (int, optional): Number of bin centers. Defaults to 64.
            bin_centers_type (str, optional): "normed" or "softplus". Activation type used for bin centers. For "normed" bin centers, linear normalization trick is applied. This results in bounded bin centers.
                                               For "softplus", softplus activation is used and thus are unbounded. Defaults to "softplus".
            bin_embedding_dim (int, optional): bin embedding dimension. Defaults to 128.
            min_depth (float, optional): Lower bound for normed bin centers. Defaults to 1e-3.
            max_depth (float, optional): Upper bound for normed bin centers. Defaults to 10.
            n_attractors (List[int], optional): Number of bin attractors at decoder layers. Defaults to [16, 8, 4, 1].
            attractor_alpha (int, optional): Proportional attractor strength. Refer to models.layers.attractor for more details. Defaults to 300.
            attractor_gamma (int, optional): Exponential attractor strength. Refer to models.layers.attractor for more details. Defaults to 2.
            attractor_kind (str, optional): Attraction aggregation "sum" or "mean". Defaults to 'sum'.
            attractor_type (str, optional): Type of attractor to use; "inv" (Inverse attractor) or "exp" (Exponential attractor). Defaults to 'exp'.
            min_temp (int, optional): Lower bound for temperature of output probability distribution. Defaults to 5.
            max_temp (int, optional): Upper bound for temperature of output probability distribution. Defaults to 50.
            train_midas (bool, optional): Whether to train "core", the base midas model. Defaults to True.
            midas_lr_factor (int, optional): Learning rate reduction factor for base midas model except its encoder and positional encodings. Defaults to 10.
            encoder_lr_factor (int, optional): Learning rate reduction factor for the encoder in midas model. Defaults to 10.
            pos_enc_lr_factor (int, optional): Learning rate reduction factor for positional encodings in the base midas model. Defaults to 10.
        """
        super().__init__()

        self.core = core
        self.max_depth = max_depth
        self.min_depth = min_depth
        self.min_temp = min_temp
        self.bin_centers_type = bin_centers_type

        self.midas_lr_factor = midas_lr_factor
        self.encoder_lr_factor = encoder_lr_factor
        self.pos_enc_lr_factor = pos_enc_lr_factor
        self.train_midas = train_midas
        self.inverse_midas = inverse_midas

        if self.encoder_lr_factor <= 0:
            self.core.freeze_encoder(
                freeze_rel_pos=self.pos_enc_lr_factor <= 0)

        N_MIDAS_OUT = 32
        btlnck_features = self.core.output_channels[0]
        num_out_features = self.core.output_channels[1:]

        self.conv2 = nn.Conv2d(btlnck_features, btlnck_features,
                               kernel_size=1, stride=1, padding=0)  # btlnck conv

        if bin_centers_type == "normed":
            SeedBinRegressorLayer = SeedBinRegressor
            Attractor = AttractorLayer
        elif bin_centers_type == "softplus":
            SeedBinRegressorLayer = SeedBinRegressorUnnormed
            Attractor = AttractorLayerUnnormed
        elif bin_centers_type == "hybrid1":
            SeedBinRegressorLayer = SeedBinRegressor
            Attractor = AttractorLayerUnnormed
        elif bin_centers_type == "hybrid2":
            SeedBinRegressorLayer = SeedBinRegressorUnnormed
            Attractor = AttractorLayer
        else:
            raise ValueError(
                "bin_centers_type should be one of 'normed', 'softplus', 'hybrid1', 'hybrid2'")

        self.seed_bin_regressor = SeedBinRegressorLayer(
            btlnck_features, n_bins=n_bins, min_depth=min_depth, max_depth=max_depth)
        self.seed_projector = Projector(btlnck_features, bin_embedding_dim)
        self.projectors = nn.ModuleList([
            Projector(num_out, bin_embedding_dim)
            for num_out in num_out_features
        ])
        self.attractors = nn.ModuleList([
            Attractor(bin_embedding_dim, n_bins, n_attractors=n_attractors[i], min_depth=min_depth, max_depth=max_depth,
                      alpha=attractor_alpha, gamma=attractor_gamma, kind=attractor_kind, attractor_type=attractor_type)
            for i in range(len(num_out_features))
        ])

        last_in = N_MIDAS_OUT + 1  # +1 for relative depth

        # use log binomial instead of softmax
        self.conditional_log_binomial = ConditionalLogBinomial(
            last_in, bin_embedding_dim, n_classes=n_bins, min_temp=min_temp, max_temp=max_temp)

    def forward(self, x, return_final_centers=False, denorm=False, return_probs=False, **kwargs):
        """
        Args:
            x (torch.Tensor): Input image tensor of shape (B, C, H, W)
            return_final_centers (bool, optional): Whether to return the final bin centers. Defaults to False.
            denorm (bool, optional): Whether to denormalize the input image. This reverses ImageNet normalization as midas normalization is different. Defaults to False.
            return_probs (bool, optional): Whether to return the output probability distribution. Defaults to False.
        
        Returns:
            dict: Dictionary containing the following keys:
                - rel_depth (torch.Tensor): Relative depth map of shape (B, H, W)
                - metric_depth (torch.Tensor): Metric depth map of shape (B, 1, H, W)
                - bin_centers (torch.Tensor): Bin centers of shape (B, n_bins). Present only if return_final_centers is True
                - probs (torch.Tensor): Output probability distribution of shape (B, n_bins, H, W). Present only if return_probs is True

        """
        b, c, h, w = x.shape
        # print("input shape ", x.shape)
        self.orig_input_width = w
        self.orig_input_height = h
        rel_depth, out = self.core(x, denorm=denorm, return_rel_depth=True)
        # print("output shapes", rel_depth.shape, out.shape)

        outconv_activation = out[0]
        btlnck = out[1]
        x_blocks = out[2:]

        x_d0 = self.conv2(btlnck)
        x = x_d0
        _, seed_b_centers = self.seed_bin_regressor(x)

        if self.bin_centers_type == 'normed' or self.bin_centers_type == 'hybrid2':
            b_prev = (seed_b_centers - self.min_depth) / \
                (self.max_depth - self.min_depth)
        else:
            b_prev = seed_b_centers

        prev_b_embedding = self.seed_projector(x)

        # unroll this loop for better performance
        for projector, attractor, x in zip(self.projectors, self.attractors, x_blocks):
            b_embedding = projector(x)
            b, b_centers = attractor(
                b_embedding, b_prev, prev_b_embedding, interpolate=True)
            b_prev = b.clone()
            prev_b_embedding = b_embedding.clone()

        last = outconv_activation

        if self.inverse_midas:
            # invert depth followed by normalization
            rel_depth = 1.0 / (rel_depth + 1e-6)
            rel_depth = (rel_depth - rel_depth.min()) / \
                (rel_depth.max() - rel_depth.min())
        # concat rel depth with last. First interpolate rel depth to last size
        rel_cond = rel_depth.unsqueeze(1)
        rel_cond = nn.functional.interpolate(
            rel_cond, size=last.shape[2:], mode='bilinear', align_corners=True)
        last = torch.cat([last, rel_cond], dim=1)

        b_embedding = nn.functional.interpolate(
            b_embedding, last.shape[-2:], mode='bilinear', align_corners=True)
        x = self.conditional_log_binomial(last, b_embedding)

        # Now depth value is Sum px * cx , where cx are bin_centers from the last bin tensor
        # print(x.shape, b_centers.shape)
        b_centers = nn.functional.interpolate(
            b_centers, x.shape[-2:], mode='bilinear', align_corners=True)
        out = torch.sum(x * b_centers, dim=1, keepdim=True)

        # Structure output dict
        output = dict(metric_depth=out)
        if return_final_centers or return_probs:
            output['bin_centers'] = b_centers

        if return_probs:
            output['probs'] = x

        return output

    def get_lr_params(self, lr):
        """
        Learning rate configuration for different layers of the model
        Args:
            lr (float) : Base learning rate
        Returns:
            list : list of parameters to optimize and their learning rates, in the format required by torch optimizers.
        """
        param_conf = []
        if self.train_midas:
            if self.encoder_lr_factor > 0:
                param_conf.append({'params': self.core.get_enc_params_except_rel_pos(
                ), 'lr': lr / self.encoder_lr_factor})

            if self.pos_enc_lr_factor > 0:
                param_conf.append(
                    {'params': self.core.get_rel_pos_params(), 'lr': lr / self.pos_enc_lr_factor})

            midas_params = self.core.core.scratch.parameters()
            midas_lr_factor = self.midas_lr_factor
            param_conf.append(
                {'params': midas_params, 'lr': lr / midas_lr_factor})

        remaining_modules = []
        for name, child in self.named_children():
            if name != 'core':
                remaining_modules.append(child)
        remaining_params = itertools.chain(
            *[child.parameters() for child in remaining_modules])

        param_conf.append({'params': remaining_params, 'lr': lr})

        return param_conf

    @staticmethod
    def build(midas_model_type="DPT_BEiT_L_384", pretrained_resource=None, use_pretrained_midas=False, train_midas=False, freeze_midas_bn=True, **kwargs):
        core = MidasCore.build(midas_model_type=midas_model_type, use_pretrained_midas=use_pretrained_midas,
                               train_midas=train_midas, fetch_features=True, freeze_bn=freeze_midas_bn, **kwargs)
        model = ZoeDepth(core, **kwargs)
        if pretrained_resource:
            assert isinstance(pretrained_resource, str), "pretrained_resource must be a string"
            model = load_state_from_resource(model, pretrained_resource)
        return model

    @staticmethod
    def build_from_config(config):
        return ZoeDepth.build(**config)



# model settings
_base_ = ['../backbones/dino_vit_giant2_reg.py']
model = dict(
    type='DensePredModel',
    decode_head=dict(
        type='RAFTDepthDPT',
        in_channels=[1536, 1536, 1536, 1536],
        use_cls_token=True,
        feature_channels = [384, 768, 1536, 1536], # [2/7, 1/7, 1/14, 1/14]
        decoder_channels = [192, 384, 768, 1536, 1536], # [4/7, 2/7, 1/7, 1/14, 1/14]
        up_scale = 7,
        hidden_channels=[192, 192, 192, 192], # [x_4, x_8, x_16, x_32] [192, 384, 768, 1536]
        n_gru_layers=3,
        n_downsample=2,
        iters=3,
        slow_fast_gru=True,
        num_register_tokens=4,
        prefix='decode_heads.'),
)



import os
import os.path as osp
import cv2
import time
import sys
CODE_SPACE=os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(CODE_SPACE)
import argparse
import mmcv
import torch
import torch.distributed as dist
import torch.multiprocessing as mp

try:
    from mmcv.utils import Config, DictAction
except:
    from mmengine import Config, DictAction
from datetime import timedelta
import random
import numpy as np
from mono.utils.logger import setup_logger
import glob
from mono.utils.comm import init_env
from mono.model.monodepth_model import get_configured_monodepth_model
from mono.utils.running import load_ckpt
from mono.utils.do_test import do_scalecano_test_with_custom_data
from mono.utils.mldb import load_data_info, reset_ckpt_path
from mono.utils.custom_data import load_from_annos, load_data

def parse_args():
    parser = argparse.ArgumentParser(description='Train a segmentor')
    parser.add_argument('config', help='train config file path')
    parser.add_argument('--show-dir', help='the dir to save logs and visualization results')
    parser.add_argument('--load-from', help='the checkpoint file to load weights from')
    parser.add_argument('--node_rank', type=int, default=0)
    parser.add_argument('--nnodes', type=int, default=1, help='number of nodes')
    parser.add_argument('--options', nargs='+', action=DictAction, help='custom options')
    parser.add_argument('--launcher', choices=['None', 'pytorch', 'slurm', 'mpi', 'ror'], default='slurm', help='job launcher')
    parser.add_argument('--test_data_path', default='None', type=str, help='the path of test data')
    parser.add_argument('--batch_size', default=1, type=int, help='the batch size for inference')
    args = parser.parse_args()
    return args

def main(args):
    os.chdir(CODE_SPACE)
    cfg = Config.fromfile(args.config)
    
    if args.options is not None:
        cfg.merge_from_dict(args.options)
        
    # show_dir is determined in this priority: CLI > segment in file > filename
    if args.show_dir is not None:
        # update configs according to CLI args if args.show_dir is not None
        cfg.show_dir = args.show_dir
    else:
        # use condig filename + timestamp as default show_dir if args.show_dir is None
        cfg.show_dir = osp.join('./show_dirs', 
                                osp.splitext(osp.basename(args.config))[0],
                                args.timestamp)
    
    # ckpt path
    if args.load_from is None:
        raise RuntimeError('Please set model path!')
    cfg.load_from = args.load_from
    cfg.batch_size = args.batch_size
    
    # load data info
    data_info = {}
    load_data_info('data_info', data_info=data_info)
    cfg.mldb_info = data_info
    # update check point info
    reset_ckpt_path(cfg.model, data_info)
    
    # create show dir
    os.makedirs(osp.abspath(cfg.show_dir), exist_ok=True)
    
    # init the logger before other steps
    cfg.log_file = osp.join(cfg.show_dir, f'{args.timestamp}.log')
    logger = setup_logger(cfg.log_file)
    
    # log some basic info
    logger.info(f'Config:\n{cfg.pretty_text}')
    
    # init distributed env dirst, since logger depends on the dist info
    if args.launcher == 'None':
        cfg.distributed = False
    else:
        cfg.distributed = True
        init_env(args.launcher, cfg)
    logger.info(f'Distributed training: {cfg.distributed}')
    
    # dump config 
    cfg.dump(osp.join(cfg.show_dir, osp.basename(args.config)))
    test_data_path = args.test_data_path
    if not os.path.isabs(test_data_path):
        test_data_path = osp.join(CODE_SPACE, test_data_path)

    if 'json' in test_data_path:
        test_data = load_from_annos(test_data_path)
    else:
        test_data = load_data(args.test_data_path)
    
    if not cfg.distributed:
        main_worker(0, cfg, args.launcher, test_data)
    else:
        # distributed training
        if args.launcher == 'ror':
            local_rank = cfg.dist_params.local_rank
            main_worker(local_rank, cfg, args.launcher, test_data)
        else:
            mp.spawn(main_worker, nprocs=cfg.dist_params.num_gpus_per_node, args=(cfg, args.launcher, test_data))
        
def main_worker(local_rank: int, cfg: dict, launcher: str, test_data: list):
    if cfg.distributed:
        cfg.dist_params.global_rank = cfg.dist_params.node_rank * cfg.dist_params.num_gpus_per_node + local_rank
        cfg.dist_params.local_rank = local_rank

        if launcher == 'ror':
            init_torch_process_group(use_hvd=False)
        else:
            torch.cuda.set_device(local_rank)
            default_timeout = timedelta(minutes=30)
            dist.init_process_group(
                backend=cfg.dist_params.backend,
                init_method=cfg.dist_params.dist_url,
                world_size=cfg.dist_params.world_size,
                rank=cfg.dist_params.global_rank,
                timeout=default_timeout)
    
    logger = setup_logger(cfg.log_file)
    # build model
    model = get_configured_monodepth_model(cfg, )
    
    # config distributed training
    if cfg.distributed:
        model = torch.nn.parallel.DistributedDataParallel(model.cuda(),
                                                          device_ids=[local_rank],
                                                          output_device=local_rank,
                                                          find_unused_parameters=True)
    else:
        model = torch.nn.DataParallel(model).cuda()
        
    # load ckpt
    model, _,  _, _ = load_ckpt(cfg.load_from, model, strict_match=False)
    model.eval()
    
    do_scalecano_test_with_custom_data(
        model, 
        cfg,
        test_data,
        logger,
        cfg.distributed,
        local_rank,
        cfg.batch_size,
    )
    
if __name__ == '__main__':
    args = parse_args()
    timestamp = time.strftime('%Y%m%d_%H%M%S', time.localtime())
    args.timestamp = timestamp
    main(args)    


import abc
from abc import ABC, abstractmethod
from typing import Any

class BaseBlender(ABC, metaclass=abc.ABCMeta):
    params: dict[str, Any]
    def __init__(self, **kwargs: Any) -> None: ...
    @abstractmethod
    def blend(self, images, masks): ...
    def update(self, **kwargs) -> None: ...



# Copyright (C) 2024-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).
#
# --------------------------------------------------------
# AachenDayNight dataloader
# --------------------------------------------------------
import os
from dust3r_visloc.datasets.base_colmap import BaseVislocColmapDataset


class VislocAachenDayNight(BaseVislocColmapDataset):
    def __init__(self, root, subscene, pairsfile, topk=1, cache_sfm=False):
        assert subscene in [None, '', 'day', 'night', 'all']
        self.subscene = subscene
        image_path = os.path.join(root, 'images')
        map_path = os.path.join(root, 'mapping/colmap/reconstruction')
        query_path = os.path.join(root, 'kapture', 'query')
        pairsfile_path = os.path.join(root, 'pairsfile/query', pairsfile + '.txt')
        super().__init__(image_path=image_path, map_path=map_path,
                         query_path=query_path, pairsfile_path=pairsfile_path,
                         topk=topk, cache_sfm=cache_sfm)
        self.scenes = [filename for filename in self.scenes if filename in self.pairs]
        if self.subscene == 'day' or self.subscene == 'night':
            self.scenes = [filename for filename in self.scenes if self.subscene in filename]



import os
import json
import torch
import torchvision.transforms as transforms
import os.path
import numpy as np
import cv2
from torch.utils.data import Dataset
import random
import mono.utils.transform as img_transform
import copy 
from mono.utils.comm import get_func
import pickle
import logging
import multiprocessing as mp
import ctypes
"""
Dataset annotations are saved in a Json file. All data, including rgb, depth, pose, and so on, captured within the same frame are saved in the same dict.
All frames are organized in a list. In each frame, it may contains the some or all of following data format. 

# Annotations for the current central RGB/depth cameras.

'rgb':          rgb image in the current frame.
'depth':        depth map in the current frame.
'sem':          semantic mask in the current frame.
'cam_in':       camera intrinsic parameters of the current rgb camera. 
'cam_ex':       camera extrinsic parameters of the current rgb camera.
'cam_ex_path':  path to the extrinsic parameters.
'pose':         pose in current frame.
'timestamp_rgb':    time stamp of current rgb image.

# Annotations for the left hand RGB/depth cameras.

'rgb_l':          rgb image of the left hand camera in the current frame.
'depth_l':        depth map of the left hand camera in the current frame.
'sem_l':          semantic mask of the left hand camera in the current frame.
'cam_in_l':       camera intrinsic parameters of the left hand rgb camera in the current frame.
'cam_ex_l':       camera extrinsic parameters of the left hand rgb camera in the current frame.
'cam_ex_path':    path to the extrinsic parameters.
'pose_l':         pose of the left hand camera  in the incurrent frame.
'timestamp_rgb_l':    time stamp of the rgb img captured by the left hand camera.

# Annotations for the right RGB/depth cameras, which is on the left hand of the current central cameras.

'rgb_r':          rgb image of the right hand camera in the current frame.
'depth_r':        depth map of the right hand camera in the current frame.
'sem_r':          semantic mask of the right hand camera in the current frame.
'cam_in_r':       camera intrinsic parameters of the right hand rgb camera in the current frame.
'cam_ex_r':       camera extrinsic parameters of the right hand rgb camera in the current frame.
'cam_ex_path_r':  path to the extrinsic parameters.
'pose_r':         pose of the right hand camera  in the incurrent frame.
'timestamp_rgb_r':    time stamp of the rgb img captured by the right hand camera.

# Annotations for the central RGB/depth cameras in the last frame.

'rgb_pre':          rgb image of the central camera in the last frame.
'depth_pre':        depth map of the central camera in the last frame.
'sem_pre':          semantic mask of the central camera in the last frame.
'cam_in_pre':       camera intrinsic parameters of the central rgb camera in the last frame.
'cam_ex_pre':       camera extrinsic parameters of the central rgb camera in the last frame.
'cam_ex_path_pre':  path to the extrinsic parameters.
'pose_pre':         pose of the central camera  in the last frame.
'timestamp_rgb_pre':    time stamp of the rgb img captured by the central camera.

# Annotations for the central RGB/depth cameras in the next frame.

'rgb_next':          rgb image of the central camera in the next frame.
'depth_next':        depth map of the central camera in the next frame.
'sem_next':          semantic mask of the central camera in the next frame.
'cam_in_next':       camera intrinsic parameters of the central rgb camera in the next frame.
'cam_ex_next':       camera extrinsic parameters of the central rgb camera in the next frame.
'cam_ex_path_next':  path to the extrinsic parameters.
'pose_next':         pose of the central camera  in the next frame.
'timestamp_rgb_next':    time stamp of the rgb img captured by the central camera.
"""

class BaseDataset(Dataset):
    def __init__(self, cfg, phase, **kwargs):
        super(BaseDataset, self).__init__()
        self.cfg = cfg
        self.phase = phase
        self.db_info = kwargs['db_info']
        
        # root dir for data
        self.data_root = os.path.join(self.db_info['db_root'], self.db_info['data_root']) 
        # depth/disp data root
        disp_root = self.db_info['disp_root'] if 'disp_root' in self.db_info else None
        self.disp_root = os.path.join(self.db_info['db_root'], disp_root) if disp_root is not None else None
        depth_root = self.db_info['depth_root'] if 'depth_root' in self.db_info else None
        self.depth_root = os.path.join(self.db_info['db_root'], depth_root) if depth_root is not None \
            else self.data_root
        # meta data root
        meta_data_root = self.db_info['meta_data_root'] if 'meta_data_root' in self.db_info else None
        self.meta_data_root = os.path.join(self.db_info['db_root'], meta_data_root) if meta_data_root is not None \
            else None
        # semantic segmentation labels root
        sem_root = self.db_info['semantic_root'] if 'semantic_root' in self.db_info else None
        self.sem_root = os.path.join(self.db_info['db_root'], sem_root) if sem_root is not None \
            else None
        # depth valid mask labels root
        depth_mask_root = self.db_info['depth_mask_root'] if 'depth_mask_root' in self.db_info else None
        self.depth_mask_root = os.path.join(self.db_info['db_root'], depth_mask_root) if depth_mask_root is not None \
            else None
        # surface normal labels root
        norm_root = self.db_info['normal_root'] if 'normal_root' in self.db_info else None
        self.norm_root = os.path.join(self.db_info['db_root'], norm_root) if norm_root is not None \
            else None
        # data annotations path
        self.data_annos_path = os.path.join(self.db_info['db_root'], self.db_info['%s_annotations_path' % phase])

        # load annotations
        self.data_info = self.load_annotations()
        whole_data_size = len(self.data_info['files']) 
        
        # sample a subset for training/validation/testing
        # such method is deprecated, each training may get different sample list
        
        cfg_sample_ratio = cfg.data[phase].sample_ratio 
        cfg_sample_size = int(cfg.data[phase].sample_size)
        self.sample_size = int(whole_data_size * cfg_sample_ratio) if cfg_sample_size == -1 \
                           else (cfg_sample_size if cfg_sample_size < whole_data_size else whole_data_size)
        random.seed(100) # set the random seed
        sample_list_of_whole_data = random.sample(list(range(whole_data_size)), self.sample_size)

        self.data_size = self.sample_size
        self.annotations = {'files': [self.data_info['files'][i] for i in sample_list_of_whole_data]}
        self.sample_list = list(range(self.data_size))
        
        # config transforms for the input and label
        self.transforms_cfg = cfg.data[phase]['pipeline']
        self.transforms_lib = 'mono.utils.transform.'

        self.img_file_type = ['.png', '.jpg', '.jpeg', '.bmp', '.tif']
        self.np_file_type = ['.npz', '.npy']

        # update canonical sparce information
        self.data_basic = copy.deepcopy(kwargs)
        canonical = self.data_basic.pop('canonical_space')
        self.data_basic.update(canonical)
        self.disp_scale = 10.0
        self.depth_range = kwargs['depth_range'] # predefined depth range for the network
        self.clip_depth_range = kwargs['clip_depth_range'] # predefined depth range for data processing
        self.depth_normalize = kwargs['depth_normalize']
        
        self.img_transforms = img_transform.Compose(self.build_data_transforms())
        self.EPS = 1e-6

        # self.tmpl_info = ['rgb_sr', 'rgb_pre', 'rgb_next']
        # self.tgt2ref_pose_lookup = {'rgb_sr': 'cam_ex', 'rgb_pre': 'pose_pre', 'rgb_next': 'pose_next'}

        # dataset info
        self.data_name = cfg.data_name
        self.data_type = cfg.data_type # there are mainly four types, i.e. ['rel', 'sfm', 'stereo', 'lidar']
        self.logger = logging.getLogger()
        self.logger.info(f'{self.data_name} in {self.phase} whole data size: {whole_data_size}')
        
        # random crop size for training
        crop_size = kwargs['crop_size']
        shared_array_base = mp.Array(ctypes.c_int32, 2)
        shared_array = np.ctypeslib.as_array(shared_array_base.get_obj())
        shared_array[0] = crop_size[0]
        shared_array[1] = crop_size[1]
        # self.random_crop_size = torch.from_numpy(np.array([0,0])) #torch.from_numpy(shared_array)
        self.random_crop_size = torch.from_numpy(shared_array)

        
    def __name__(self):
        return self.data_name

    def __len__(self):
        return self.data_size

    def load_annotations(self):
        if not os.path.exists(self.data_annos_path):
            raise RuntimeError(f'Cannot find {self.data_annos_path} annotations.')

        with open(self.data_annos_path, 'r')  as f:
            annos = json.load(f)
        return annos
    
    def build_data_transforms(self):
        transforms_list = []
        for transform in self.transforms_cfg:
            args = copy.deepcopy(transform)
            # insert the canonical space configs
            args.update(self.data_basic)
      
            obj_name = args.pop('type')
            obj_path = self.transforms_lib + obj_name
            obj_cls = get_func(obj_path)
            
            obj = obj_cls(**args)
            transforms_list.append(obj)
        return transforms_list
    
        
    def load_data(self, path: str, is_rgb_img: bool=False):
        if not os.path.exists(path):
            self.logger.info(f'>>>>{path} does not exist.')
            # raise RuntimeError(f'{path} does not exist.')

        data_type = os.path.splitext(path)[-1]
        if data_type in self.img_file_type:
            if is_rgb_img:
                data = cv2.imread(path)
            else:
                data = cv2.imread(path, -1)
        elif data_type in self.np_file_type:
            data = np.load(path)
        else:
            raise RuntimeError(f'{data_type} is not supported in current version.')
        
        try:
            return data.squeeze()
        except:
            temp = 1
            raise RuntimeError(f'{path} is not successfully loaded.')
    
    def __getitem__(self, idx: int) -> dict:
        if self.phase == 'test':
            return self.get_data_for_test(idx)
        else:
            return self.get_data_for_trainval(idx)

    def get_data_for_trainval(self, idx: int):
        anno = self.annotations['files'][idx]
        meta_data = self.load_meta_data(anno)
        
        data_path = self.load_data_path(meta_data)
        data_batch = self.load_batch(meta_data, data_path)
        # if data_path['sem_path'] is not None:
        #     print(self.data_name)

        curr_rgb, curr_depth, curr_normal, curr_sem, curr_cam_model = data_batch['curr_rgb'], data_batch['curr_depth'], data_batch['curr_normal'], data_batch['curr_sem'], data_batch['curr_cam_model']
        #curr_stereo_depth = data_batch['curr_stereo_depth']
        
        # A patch for stereo depth dataloader (no need to modify specific datasets)
        if 'curr_stereo_depth' in data_batch.keys():
            curr_stereo_depth = data_batch['curr_stereo_depth']
        else:
            curr_stereo_depth = self.load_stereo_depth_label(None, H=curr_rgb.shape[0], W=curr_rgb.shape[1]) 

        curr_intrinsic = meta_data['cam_in']
        # data augmentation
        transform_paras = dict(random_crop_size = self.random_crop_size) # dict() 
        assert curr_rgb.shape[:2] == curr_depth.shape == curr_normal.shape[:2] == curr_sem.shape
        rgbs, depths, intrinsics, cam_models, normals, other_labels, transform_paras = self.img_transforms(
                                                                   images=[curr_rgb, ], 
                                                                   labels=[curr_depth, ], 
                                                                   intrinsics=[curr_intrinsic,], 
                                                                   cam_models=[curr_cam_model, ],
                                                                   normals = [curr_normal, ],
                                                                   other_labels=[curr_sem, curr_stereo_depth],
                                                                   transform_paras=transform_paras)
        # process sky masks
        sem_mask = other_labels[0].int()
        # clip depth map 
        depth_out = self.normalize_depth(depths[0])
        # set the depth of sky region to the invalid
        depth_out[sem_mask==142] = -1 # self.depth_normalize[1] - 1e-6
        # get inverse depth
        inv_depth = self.depth2invdepth(depth_out, sem_mask==142)
        filename = os.path.basename(meta_data['rgb'])[:-4] + '.jpg'
        curr_intrinsic_mat = self.intrinsics_list2mat(intrinsics[0])
        cam_models_stacks = [
            torch.nn.functional.interpolate(cam_models[0][None, :, :, :], size=(cam_models[0].shape[1]//i, cam_models[0].shape[2]//i), mode='bilinear', align_corners=False).squeeze()
            for i in [2, 4, 8, 16, 32] 
            ]

        # stereo_depth
        if 'label_scale_factor' not in transform_paras.keys():
            transform_paras['label_scale_factor'] = 1
        stereo_depth_pre_trans = other_labels[1] * (other_labels[1] > 0.3) * (other_labels[1] < 200)
        stereo_depth = stereo_depth_pre_trans * transform_paras['label_scale_factor']
        stereo_depth = self.normalize_depth(stereo_depth)

        pad = transform_paras['pad'] if 'pad' in transform_paras else [0,0,0,0]        
        data = dict(input=rgbs[0],
                    target=depth_out,
                    intrinsic=curr_intrinsic_mat,
                    filename=filename,
                    dataset=self.data_name,
                    cam_model=cam_models_stacks,
                    pad=torch.tensor(pad),
                    data_type=[self.data_type, ],
                    sem_mask=sem_mask.int(),
                    stereo_depth= stereo_depth,
                    normal=normals[0],
                    inv_depth=inv_depth,
                    scale=transform_paras['label_scale_factor'])
        return data

    def get_data_for_test(self, idx: int):
        anno = self.annotations['files'][idx]
        meta_data = self.load_meta_data(anno)
        data_path = self.load_data_path(meta_data)
        data_batch = self.load_batch(meta_data, data_path)
        # load data
        curr_rgb, curr_depth, curr_normal, curr_cam_model = data_batch['curr_rgb'], data_batch['curr_depth'], data_batch['curr_normal'], data_batch['curr_cam_model']
        ori_curr_intrinsic = meta_data['cam_in']

        # get crop size
        transform_paras = dict()
        rgbs, depths, intrinsics, cam_models, _, other_labels, transform_paras = self.img_transforms(
                                                                   images=[curr_rgb,],  #+ tmpl_rgbs, 
                                                                   labels=[curr_depth, ], 
                                                                   intrinsics=[ori_curr_intrinsic, ], # * (len(tmpl_rgbs) + 1), 
                                                                   cam_models=[curr_cam_model, ],
                                                                   transform_paras=transform_paras)
        # depth in original size and orignial metric***
        depth_out = self.clip_depth(curr_depth) * self.depth_range[1] # self.clip_depth(depths[0]) #
        inv_depth = self.depth2invdepth(depth_out, np.zeros_like(depth_out, dtype=np.bool))
        filename = os.path.basename(meta_data['rgb'])[:-4] + '.jpg'
        curr_intrinsic_mat = self.intrinsics_list2mat(intrinsics[0])
        ori_curr_intrinsic_mat = self.intrinsics_list2mat(ori_curr_intrinsic)

        pad = transform_paras['pad'] if 'pad' in transform_paras else [0,0,0,0]
        scale_ratio = transform_paras['label_scale_factor'] if 'label_scale_factor' in transform_paras else 1.0
        cam_models_stacks = [
            torch.nn.functional.interpolate(cam_models[0][None, :, :, :], size=(cam_models[0].shape[1]//i, cam_models[0].shape[2]//i), mode='bilinear', align_corners=False).squeeze()
            for i in [2, 4, 8, 16, 32] 
            ]    
        raw_rgb = torch.from_numpy(curr_rgb)
        curr_normal = torch.from_numpy(curr_normal.transpose((2,0,1)))
        

        data = dict(input=rgbs[0],
                    target=depth_out,
                    intrinsic=curr_intrinsic_mat,
                    filename=filename,
                    dataset=self.data_name,
                    cam_model=cam_models_stacks,
                    pad=pad,
                    scale=scale_ratio,
                    raw_rgb=raw_rgb,
                    sample_id=idx,
                    data_path=meta_data['rgb'],
                    inv_depth=inv_depth,
                    normal=curr_normal,
                    )
        return data
    
    def load_data_path(self, meta_data):
        curr_rgb_path = os.path.join(self.data_root, meta_data['rgb'])
        curr_depth_path = os.path.join(self.depth_root, meta_data['depth'])
        curr_sem_path = os.path.join(self.sem_root, meta_data['sem']) \
            if self.sem_root is not None and ('sem' in meta_data) and (meta_data['sem'] is not None)  \
            else None
        # matterport3d separates xyz into three images
        if ('normal' in meta_data) and (meta_data['normal'] is not None) and (self.norm_root is not None):
            if isinstance(meta_data['normal'], dict):
                curr_norm_path = {}
                for k,v in meta_data['normal'].items():
                    curr_norm_path[k] = os.path.join(self.norm_root, v)
            else:
                curr_norm_path = os.path.join(self.norm_root, meta_data['normal'])
        else:
            curr_norm_path = None
        curr_depth_mask_path = os.path.join(self.depth_mask_root, meta_data['depth_mask']) \
            if self.depth_mask_root is not None and ('depth_mask' in meta_data) and (meta_data['depth_mask'] is not None)  \
            else None

        if ('disp' in meta_data) and (meta_data['disp'] is not None) and (self.disp_root is not None):
            if isinstance(meta_data['disp'], dict):
                curr_disp_path = {}
                for k,v in meta_data['disp'].items():
                    curr_disp_path[k] = os.path.join(self.disp_root, v)
            else:
                curr_disp_path = os.path.join(self.disp_root, meta_data['disp'])
        else:
            curr_disp_path = None

        data_path=dict(
            rgb_path=curr_rgb_path,
            depth_path=curr_depth_path,
            sem_path=curr_sem_path,
            normal_path=curr_norm_path,
            disp_path=curr_disp_path,
            depth_mask_path=curr_depth_mask_path,
            )
        return data_path
    
    def load_batch(self, meta_data, data_path):
        curr_intrinsic = meta_data['cam_in']
        # load rgb/depth
        curr_rgb, curr_depth = self.load_rgb_depth(data_path['rgb_path'], data_path['depth_path'])
        # get semantic labels
        curr_sem = self.load_sem_label(data_path['sem_path'], curr_depth)
        # create camera model
        curr_cam_model = self.create_cam_model(curr_rgb.shape[0], curr_rgb.shape[1], curr_intrinsic)       
        # get normal labels
        curr_normal = self.load_norm_label(data_path['normal_path'], H=curr_rgb.shape[0], W=curr_rgb.shape[1]) 
        # get depth mask
        depth_mask = self.load_depth_valid_mask(data_path['depth_mask_path'])
        curr_depth[~depth_mask] = -1
        # get stereo depth
        curr_stereo_depth = self.load_stereo_depth_label(data_path['disp_path'], H=curr_rgb.shape[0], W=curr_rgb.shape[1]) 

        data_batch = dict(
            curr_rgb = curr_rgb,
            curr_depth = curr_depth,
            curr_sem = curr_sem,
            curr_normal = curr_normal,
            curr_cam_model=curr_cam_model,
            curr_stereo_depth=curr_stereo_depth,
        )
        return data_batch


    def clip_depth(self, depth: np.array) -> np.array:
        depth[(depth>self.clip_depth_range[1]) | (depth<self.clip_depth_range[0])] = -1
        depth /= self.depth_range[1]
        depth[depth<self.EPS] = -1
        return depth
    
    def normalize_depth(self, depth: np.array) -> np.array:
        depth /= self.depth_range[1]
        depth[depth<self.EPS] = -1
        return depth
    
    def process_depth(self, depth: np.array, rgb:np.array=None):
        return depth
    
    def create_cam_model(self, H : int, W : int, intrinsics : list) -> np.array:
        """
        Encode the camera model (focal length and principle point) to a 4-channel map. 
        """
        fx, fy, u0, v0 = intrinsics
        f = (fx + fy) / 2.0
        # principle point location
        x_row = np.arange(0, W).astype(np.float32)
        x_row_center_norm = (x_row - u0) / W
        x_center = np.tile(x_row_center_norm, (H, 1)) # [H, W]

        y_col = np.arange(0, H).astype(np.float32) 
        y_col_center_norm = (y_col - v0) / H
        y_center = np.tile(y_col_center_norm, (W, 1)).T

        # FoV
        fov_x = np.arctan(x_center / (f / W))
        fov_y =  np.arctan(y_center/ (f / H))

        cam_model = np.stack([x_center, y_center, fov_x, fov_y], axis=2)
        return cam_model
    
    def check_data(self, data_dict : dict):
        for k, v in data_dict.items():
            if v is None:
                # print(f'{self.data_name}, {k} cannot be read!')
                self.logger.info(f'{self.data_name}, {k} cannot be read!')
    
    def intrinsics_list2mat(self, intrinsics: torch.tensor) -> torch.tensor:
        """
        Create camera intrinsic matrix.
        Args:
            intrinsics (torch.tensor, [4,]): list of camera intrinsic parameters.
        returns:
            intrinsics_mat (torch.tensor, [3x3]): camera intrinsic parameters matrix.
        """
        intrinsics_mat = torch.zeros((3,3)).float()
        intrinsics_mat[0, 0] = intrinsics[0]
        intrinsics_mat[1, 1] = intrinsics[1]
        intrinsics_mat[0, 2] = intrinsics[2]
        intrinsics_mat[1, 2] = intrinsics[3]
        intrinsics_mat[2, 2] = 1.0
        return intrinsics_mat
        
    # def load_tmpl_image(self, curr_rgb: np.array, meta_data: dict) -> dict:
    #     """
    #     Load  consecutive RGB frames.
    #     Args:
    #         anno: the annotation for this group.
    #         curr_rgb: rgb image of the current frame.
    #         meta_data: meta data information.
    #     Returns:
    #         tmpl_annos: temporal rgbs.
    #     """
    #     w_tmpl = False
        
    #     tmpl_list = []
    #     # organize temporal annotations 
    #     for i in self.tmpl_info:
    #         if (i in meta_data) and (meta_data[i] is not None) and os.path.exists(os.path.join(self.data_root, meta_data[i])):
    #             tmpl_list.append(os.path.join(self.data_root, meta_data[i]))

    #     if len(tmpl_list) == 0:
    #         rgb_tmpl = curr_rgb.copy()
    #     else:
    #         id = np.random.randint(len(tmpl_list))
    #         rgb_tmpl = self.load_data(tmpl_list[id], is_rgb_img=True)
    #         w_tmpl = True

    #     tmpl_annos = dict(
    #         tmpl_rgb_list = [rgb_tmpl,],
    #         w_tmpl = w_tmpl
    #     )
    #     return tmpl_annos

    def load_meta_data(self, anno: dict) -> dict:
        """
        Load meta data information.
        """
        if self.meta_data_root is not None and ('meta_data' in anno or 'meta' in anno):
            meta_data_path = os.path.join(self.meta_data_root, anno['meta_data']) if 'meta_data' in anno else os.path.join(self.meta_data_root, anno['meta'])
            with open(meta_data_path, 'rb') as f:
                meta_data = pickle.load(f)
            meta_data.update(anno)
        else:
            meta_data = anno
        return meta_data
    
    def load_rgb_depth(self, rgb_path: str, depth_path: str):
        """
        Load the rgb and depth map with the paths.
        """
        rgb = self.load_data(rgb_path, is_rgb_img=True)
        if rgb is None:
            self.logger.info(f'>>>>{rgb_path} has errors.')
       
        depth = self.load_data(depth_path)
        if depth is None:
            self.logger.info(f'{depth_path} has errors.')
        
        # self.check_data(dict(
        #     rgb_path=rgb,
        #     depth_path=depth,
        # ))
        depth = depth.astype(np.float)
        # if depth.shape != rgb.shape[:2]:
        #     print(f'no-equal in {self.data_name}')
        #     depth = cv2.resize(depth, rgb.shape[::-1][1:])
        
        depth  = self.process_depth(depth, rgb)
        return rgb, depth
    
    def load_sem_label(self, sem_path, depth=None, sky_id=142) -> np.array:
        H, W = depth.shape
        # if sem_path is not None:
        #     print(self.data_name)
        sem_label = cv2.imread(sem_path, 0) if sem_path is not None \
            else np.ones((H, W), dtype=np.int) * -1
        if sem_label is None:
            sem_label = np.ones((H, W), dtype=np.int) * -1
        # set dtype to int before 
        sem_label = sem_label.astype(np.int) 
        sem_label[sem_label==255] = -1
        
        # mask invalid sky region
        mask_depth_valid = depth > 1e-8
        invalid_sky_region = (sem_label==142) & (mask_depth_valid)
        if self.data_type in ['lidar', 'sfm', 'denselidar', 'denselidar_nometric']:
            sem_label[invalid_sky_region] = -1
        return sem_label
    
    def load_depth_valid_mask(self, depth_mask_path, depth=None) -> np.array:
        if depth_mask_path is None:
            return np.ones_like(depth, dtype=np.bool)
        data_type = os.path.splitext(depth_mask_path)[-1]
        if data_type in self.img_file_type:
            data = cv2.imread(depth_mask_path, -1)
        elif data_type in self.np_file_type:
            data = np.load(depth_mask_path)
        else:
            raise RuntimeError(f'{data_type} is not supported in current version.')
        data = data.astype(np.bool)
        return data
        
    def load_norm_label(self, norm_path, H, W):
        norm_gt = np.zeros((H, W, 3)).astype(np.float32)
        return norm_gt

    def load_stereo_depth_label(self, disp_path, H, W):
        stereo_depth_gt = np.zeros((H, W, 1)).astype(np.float32)
        return stereo_depth_gt

    def depth2invdepth(self, depth, sky_mask):
        inv_depth = 1.0 / depth * self.disp_scale
        inv_depth[depth<1e-6] = -1.0
        inv_depth[inv_depth < 0] = -1.0
        inv_depth[sky_mask] = 0
        return inv_depth


    def set_random_crop_size(self, random_crop_size):
        self.random_crop_size[0] = random_crop_size[0]
        self.random_crop_size[1] = random_crop_size[1]



# edge_aware_depth_losses.py
"""
Re‑implementation of the depth‑refinement loss suite, focused on:
────────────────────────────────────────────────────────────────────
• robust gradient loss (Charbonnier + edge weighting)
• percentile‑based thin‑structure loss with 1‑px dilation
• adaptive residual gate (per‑batch median)
• optional 3‑D linear / log‑linear L₁ in point space
• heteroscedastic NLL, surface–normal loss, smoothness, etc.

All routines are torch‑script friendly and drop‑in compatible with
existing *FixedDepthLoss* call‑sites.
"""

from __future__ import annotations

import math
import os
import uuid
from typing import Tuple, Dict, Any, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
import matplotlib.pyplot as plt

# ────────────────────────────────────────────────────────────────────────────────
#   helpers: image un‑normalisation & cached grids
# ────────────────────────────────────────────────────────────────────────────────

_MEAN = torch.tensor([0.485, 0.456, 0.406]).view(1, 3, 1, 1)
_STD  = torch.tensor([0.229, 0.224, 0.225]).view(1, 3, 1, 1)

def unnormalize_rgb(img: torch.Tensor) -> torch.Tensor:
    """(B,3,H,W) in ‑1…1 → 0…1 float."""
    return img * _STD.to(img.device) + _MEAN.to(img.device)

# cached normalised xy‑grids  (device,H,W) → (u,v)
_grid_cache: Dict[Tuple[torch.device, int, int], Tuple[torch.Tensor, torch.Tensor]] = {}

def get_uv_grid(h: int, w: int, device: torch.device) -> Tuple[torch.Tensor, torch.Tensor]:
    key = (device, h, w)
    if key not in _grid_cache:
        u = torch.linspace(-1, 1, w, device=device)
        v = torch.linspace(-1, 1, h, device=device)
        gv, gu = torch.meshgrid(v, u, indexing="ij")
        _grid_cache[key] = (gu.unsqueeze(0).unsqueeze(0),  # (1,1,H,W)
                             gv.unsqueeze(0).unsqueeze(0))
    return _grid_cache[key]

# ────────────────────────────────────────────────────────────────────────────────
#   robust gradient loss (Charbonnier, edge‑weighted)
# ────────────────────────────────────────────────────────────────────────────────

def _charb(x: torch.Tensor, eps: float = 1e-3) -> torch.Tensor:
    return torch.sqrt(x * x + eps * eps)

def robust_gradient_loss(
    pred: torch.Tensor,
    tgt: torch.Tensor,
    mask: torch.Tensor,
    k_edge: float = 3.0
) -> torch.Tensor:
    
    # log-domain
    pred_log = torch.log(pred + 1e-6)
    tgt_log  = torch.log(tgt  + 1e-6)

    # spatial gradients
    dx_p = pred_log[..., 1:, :] - pred_log[..., :-1, :]
    dy_p = pred_log[..., :, 1:] - pred_log[..., :, :-1]
    dx_t = tgt_log[..., 1:, :] - tgt_log[..., :-1, :]
    dy_t = tgt_log[..., :, 1:] - tgt_log[..., :, :-1]

    # valid‐pair mask for each gradient map
    dx_m = mask[..., 1:, :] * mask[..., :-1, :]
    dy_m = mask[..., :, 1:] * mask[..., :, :-1]

    # edge weights separately for dx and dy
    edge_w_x = 1.0 + k_edge * dx_t.abs()
    edge_w_y = 1.0 + k_edge * dy_t.abs()

    # Charbonnier helper
    def _charb(x, eps=1e-3):
        return torch.sqrt(x * x + eps * eps)

    # horizontal + vertical losses
    loss_x = (_charb(dx_p - dx_t) * edge_w_x * dx_m).mean()
    loss_y = (_charb(dy_p - dy_t) * edge_w_y * dy_m).mean()

    return 0.5 * (loss_x + loss_y)

# ────────────────────────────────────────────────────────────────────────────────
#   thin‑structure loss (percentile threshold + dilation)
# ────────────────────────────────────────────────────────────────────────────────

def _gradient_mag(z: torch.Tensor) -> torch.Tensor:
    gx = z[..., 1:, :] - z[..., :-1, :]
    gy = z[..., :, 1:] - z[..., :, :-1]
    g  = torch.zeros_like(z)
    g[..., 1:,  :] += gx.abs()
    g[..., :-1, :] += gx.abs()
    g[..., :, 1:] += gy.abs()
    g[..., :, :-1] += gy.abs()
    return 0.5 * g  # simple average of two contributions

def thin_structure_loss(pred: torch.Tensor, tgt: torch.Tensor, mask: torch.Tensor,
                        pct: float = .70, debug: bool = False) -> torch.Tensor:
    grad_t = _gradient_mag(tgt)
    grad_p = _gradient_mag(pred)

    thr = torch.quantile(grad_t[mask.bool()], pct)
    thin = (grad_t > thr).float()
    thin = F.max_pool2d(thin, 3, 1, 1)           # 1‑px dilate

    diff  = (grad_p - grad_t).abs()
    valid = thin * mask
    if valid.sum() < 1:
        return torch.tensor(0., device=pred.device)
    loss = (diff * valid).sum() / valid.sum()

    if debug:
        with torch.no_grad():
            plt.figure(figsize=(6,3))
            plt.subplot(131); plt.imshow(grad_t[0,0].cpu(), cmap='gray'); plt.title('∇Z target'); plt.axis('off')
            plt.subplot(132); plt.imshow(thin[0,0].cpu(), cmap='gray'); plt.title('thin mask'); plt.axis('off')
            plt.subplot(133); plt.imshow(diff[0,0].cpu()*valid[0,0].cpu(), cmap='hot'); plt.title('error×mask'); plt.axis('off')
            plt.tight_layout(); plt.show()
    return loss

# ────────────────────────────────────────────────────────────────────────────────
#   curvature‑based attention (optionally inverted)
# ────────────────────────────────────────────────────────────────────────────────

def curvature_attention(depth: torch.Tensor, mask: torch.Tensor, inverse: bool = False,
                        eps: float = 1e-6, k: float = 10.) -> torch.Tensor:
    lap = F.conv2d(depth, depth.new_tensor([[0,1,0],[1,-4,1],[0,1,0]]).view(1,1,3,3), padding=1).abs()
    flat = lap[mask.bool()]
    maxv = torch.quantile(flat, 0.99)
    norm = (lap / (maxv + eps)).clamp(0, 1)
    att  = (1. - norm) if inverse else norm
    return att

# ────────────────────────────────────────────────────────────────────────────────
#   point‑space utilities
# ────────────────────────────────────────────────────────────────────────────────

def depth_to_points(depth: torch.Tensor) -> torch.Tensor:
    B, _, H, W = depth.shape
    u, v = get_uv_grid(H, W, depth.device)
    return torch.cat([depth * u, depth * v, depth], dim=1)  # (B,3,H,W)

# ────────────────────────────────────────────────────────────────────────────────
#   FixedDepthLoss class
# ────────────────────────────────────────────────────────────────────────────────

# class FixedDepthLoss(nn.Module):
#     def __init__(self, *,
#                  silog_weight=1., grad_weight=1., smooth_weight=1.,
#                  thin_weight=.0, normal_weight=.0,
#                  l1_weight=.0, l1_3d_mode="none", use_smooth_l1=True,
#                  silog_lambda=.5, k_edge_grad=3.0, thin_pct=.70,
#                  residual_reg_weight=.02, residual_tau_mode="adaptive",
#                  hetero=False, nll_weight=1., nll_ramp_epochs=5,
#                  use_mask_bce=False,            # ← already there
#                  mask_weight=0.2,               # ← NEW  (default from yaml)
#                  sixth_weight=.2):
#         super().__init__()

#         # ---------------- save everything ----------------
#         self.silog      = SiLogLoss(lambd=silog_lambda)

#         # weights
#         self.silog_w, self.grad_w, self.smooth_w = silog_weight, grad_weight, smooth_weight
#         self.thin_w,  self.normal_w, self.l1_w   = thin_weight,  normal_weight, l1_weight

#         # flags & hyper-params
#         self.l1_3d_mode     = l1_3d_mode
#         self.use_smooth_l1  = use_smooth_l1
#         self.k_edge_grad    = k_edge_grad
#         self.thin_pct       = thin_pct
#         self.residual_w     = residual_reg_weight
#         self.res_tau_mode   = residual_tau_mode
#         self.hetero, self.nll_w, self.nll_epochs = hetero, nll_weight, nll_ramp_epochs
#         self.sixth_w        = sixth_weight
#         self.use_mask_bce   = use_mask_bce
#         self.mask_weight    = mask_weight          # ← store it!

#     # ---------------------------------------------------------------------
#     def forward(self, pred, tgt, mask, *, pred_mask: torch.Tensor | None = None, extra=None, teacher=None, epoch: int = 0, image=None):
#         mask_f = mask.float(); mask_sum = mask_f.sum().clamp(min=1.)
#         loss, logs = 0., {}

#         # ---------- heteroscedastic NLL ----------
#         if self.hetero and extra is not None:
#             logv = extra.clamp_(-8., 4.)
#             invv = torch.exp(-logv)
#             nll  = 0.5 * ((pred - tgt) ** 2 * invv + logv + math.log(2*math.pi))
#             nll  = (nll * mask_f).sum() / mask_sum
#             ramp = min(1., epoch / float(self.nll_epochs))
#             loss += ramp * self.nll_w * nll;   logs['nll'] = nll.item()

#         # ---------- residual regulariser ----------
#         if teacher is not None and self.residual_w > 0:
#             if teacher.ndim == 3: teacher = teacher.unsqueeze(1)
#             if teacher.shape != pred.shape:
#                 teacher = F.interpolate(teacher, pred.shape[-2:], mode='bilinear', align_corners=False)
#             res  = pred - teacher.detach()
#             errc = (teacher - tgt).abs()
#             tau  = 0.5 * torch.quantile(errc[mask.bool()], 0.5) if self.res_tau_mode == 'adaptive' else 0.20
#             gate = (errc < tau).float()
#             Lres = ((res ** 2) * gate * mask_f).sum() / (gate*mask_f).sum().clamp(min=1.)
#             loss += self.residual_w * Lres; logs['residual_reg'] = Lres.item()
#             logs['gate_ratio'] = gate.mean().item()

#         # ---------- SiLog ----------
#         if self.silog_w:
#             sil = self.silog(pred, tgt, mask); loss += self.silog_w * sil; logs['silog'] = sil.item()

#         # ---------- L1 (depth or 3‑D) ----------
#         if self.l1_w:
#             if self.l1_3d_mode != 'none':
#                 pt_p, pt_t = depth_to_points(pred), depth_to_points(tgt)
#                 if self.l1_3d_mode == 'log':
#                     pt_p, pt_t = torch.log(pt_p.clamp_min(1e-6)), torch.log(pt_t.clamp_min(1e-6))
#                 diff = (pt_p - pt_t).abs() * mask_f
#                 per_item = diff.view(diff.size(0), -1).mean(1)
#                 if self.l1_3d_mode == 'log' and self.sixth_w != 1.:
#                     w = torch.ones_like(per_item); w[5::6] = self.sixth_w; per_item *= w
#                 l1 = per_item.mean()
#             else:
#                 beta = .15
#                 abs_e = (pred - tgt).abs()
#                 if self.use_smooth_l1:
#                     l1_map = torch.where(abs_e < beta, 0.5*abs_e*abs_e/beta, abs_e-0.5*beta)
#                 else:
#                     l1_map = abs_e
#                 l1 = (l1_map * mask_f).sum() / mask_sum
#             loss += self.l1_w * l1; logs['l1'] = l1.item()

#         # ---------- gradient & thin losses ----------
#         if self.grad_w:
#             g = robust_gradient_loss(pred, tgt, mask, k_edge=self.k_edge_grad)
#             loss += self.grad_w * g; logs['grad'] = g.item()
#         if self.thin_w:
#             t = thin_structure_loss(pred, tgt, mask, pct=self.thin_pct)
#             loss += self.thin_w * t; logs['thin'] = t.item()

#         # inside FixedDepthLoss.forward, where you do:
#         if self.use_mask_bce and pred_mask is not None:
#             # pred_mask is raw logits, mask is {0,1}-valued GT where 1=valid
#             mask_f = mask.float()

#             # invert the GT so that 1→0 and 0→1
#             inv_mask = 1.0 - mask_f

#             bce = F.binary_cross_entropy_with_logits(
#                 pred_mask,         # (B,1,H,W) logits
#                 inv_mask,          # now 1 for invalid, 0 for valid
#                 reduction="mean"
#             )
#             loss += self.mask_weight * bce
#             logs["mask_bce"] = bce.item()

#         # (optional) smoothness & normals can be re‑added here …
#         logs['total'] = loss.item(); return loss, logs

import math
import torch
import torch.nn as nn
import torch.nn.functional as F

class FixedDepthLoss(nn.Module):
    def __init__(
        self,
        *,
        silog_weight=1., grad_weight=1., smooth_weight=1.,
        thin_weight=0., normal_weight=0.,
        l1_weight=0., l1_3d_mode="none", use_smooth_l1=True,
        silog_lambda=0.5, k_edge_grad=3.0, thin_pct=0.70,
        residual_reg_weight=0.02, residual_tau_mode="adaptive",
        hetero=False, nll_weight=1., nll_ramp_epochs=5,
        use_mask_bce=False, mask_weight=0.2,
        sixth_weight=0.2,
        # scheduler arguments
        schedule_keys: list[str] = None,
        warmup_epochs: int = 0,
        ramp_epochs: int = 0,
    ):
        super().__init__()

        # ──────────────── save everything ────────────────
        self.silog = SiLogLoss(lambd=silog_lambda)

        # weights
        self.silog_w, self.grad_w, self.smooth_w = silog_weight, grad_weight, smooth_weight
        self.thin_w, self.normal_w, self.l1_w = thin_weight, normal_weight, l1_weight

        # flags & hyper-params
        self.l1_3d_mode = l1_3d_mode
        self.use_smooth_l1 = use_smooth_l1
        self.k_edge_grad = k_edge_grad
        self.thin_pct = thin_pct
        self.residual_w = residual_reg_weight
        self.res_tau_mode = residual_tau_mode
        self.hetero, self.nll_w, self.nll_epochs = hetero, nll_weight, nll_ramp_epochs
        self.sixth_w = sixth_weight
        self.use_mask_bce = use_mask_bce
        self.mask_weight = mask_weight

        # ───────────── schedule config ─────────────
        self.schedule_keys = schedule_keys or []
        self.warmup_epochs = warmup_epochs
        self.ramp_epochs = ramp_epochs
        # capture original values for scheduled keys
        self._orig_weights = {k: getattr(self, k) for k in self.schedule_keys}

    def schedule(self, epoch: int):
        """
        Zero out or ramp specified weights for the first warmup_epochs,
        then optionally ramp over ramp_epochs.
        """
        if epoch < self.warmup_epochs:
            scale = 0.0
        elif self.ramp_epochs > 0 and epoch < self.warmup_epochs + self.ramp_epochs:
            scale = (epoch - self.warmup_epochs) / float(self.ramp_epochs)
        else:
            scale = 1.0
        for key, original in self._orig_weights.items():
            setattr(self, key, scale * original)

    def forward(
        self,
        pred: torch.Tensor,
        tgt: torch.Tensor,
        mask: torch.Tensor,
        *,
        pred_mask: torch.Tensor | None = None,
        extra=None,
        teacher=None,
        epoch: int = 0,
        image=None,
    ):
        # apply warmup/ramp schedule at the start
        self.schedule(epoch)
        #  # ─── DEBUG PRINTS ────────────────────────────────────────────────
        # print(f"[Loss Debug] pred={tuple(pred.shape)}, tgt={tuple(tgt.shape)}, mask={tuple(mask.shape)}")
        # if isinstance(extra, torch.Tensor):
        #     print(f"[Loss Debug] extra(log_var)={tuple(extra.shape)}")
        # if isinstance(pred_mask, torch.Tensor):
        #     print(f"[Loss Debug] pred_mask={tuple(pred_mask.shape)}")
        

        mask_f = mask.float(); mask_sum = mask_f.sum().clamp(min=1.)
        loss, logs = 0.0, {}

        # ---------- heteroscedastic NLL ----------
        if self.hetero and extra is not None:
            logv = extra.clamp_(-8., 4.)
            invv = torch.exp(-logv)
            nll = 0.5 * ((pred - tgt) ** 2 * invv + logv + math.log(2 * math.pi))
            nll = (nll * mask_f).sum() / mask_sum
            ramp = min(1.0, epoch / float(self.nll_epochs))
            loss += ramp * self.nll_w * nll; logs['nll'] = nll.item()

        # ---------- residual regulariser ----------
        if teacher is not None and self.residual_w > 0:
            if teacher.ndim == 3:
                teacher = teacher.unsqueeze(1)
            if teacher.shape != pred.shape:
                teacher = F.interpolate(teacher, pred.shape[-2:], mode='bilinear', align_corners=False)
            res = pred - teacher.detach()
            errc = (teacher - tgt).abs()
            if self.res_tau_mode == 'adaptive':
                tau = 0.5 * torch.quantile(errc[mask.bool()], 0.5)
            else:
                tau = 0.20
            gate = (errc < tau).float()
            Lres = ((res ** 2) * gate * mask_f).sum() / (gate * mask_f).sum().clamp(min=1.)
            loss += self.residual_w * Lres; logs['residual_reg'] = Lres.item()
            logs['gate_ratio'] = gate.mean().item()

        # ---------- SiLog ----------
        if self.silog_w:
            sil = self.silog(pred, tgt, mask); loss += self.silog_w * sil; logs['silog'] = sil.item()

        # ---------- L1 (depth or 3D) ----------
        if self.l1_w:
            if self.l1_3d_mode != 'none':
                pt_p, pt_t = depth_to_points(pred), depth_to_points(tgt)
                if self.l1_3d_mode == 'log':
                    pt_p = torch.log(pt_p.clamp_min(1e-6)); pt_t = torch.log(pt_t.clamp_min(1e-6))
                diff = (pt_p - pt_t).abs() * mask_f
                per_item = diff.view(diff.size(0), -1).mean(1)
                if self.l1_3d_mode == 'log' and self.sixth_w != 1.0:
                    w = torch.ones_like(per_item); w[5::6] = self.sixth_w; per_item *= w
                l1 = per_item.mean()
            else:
                beta = 0.15
                abs_e = (pred - tgt).abs()
                if self.use_smooth_l1:
                    l1_map = torch.where(abs_e < beta,
                                         0.5 * abs_e * abs_e / beta,
                                         abs_e - 0.5 * beta)
                else:
                    l1_map = abs_e
                l1 = (l1_map * mask_f).sum() / mask_sum
            loss += self.l1_w * l1; logs['l1'] = l1.item()

        # ---------- gradient & thin losses ----------
        if self.grad_w:
            g = robust_gradient_loss(pred, tgt, mask, k_edge=self.k_edge_grad)
            loss += self.grad_w * g; logs['grad'] = g.item()
        if self.thin_w:
            t = thin_structure_loss(pred, tgt, mask, pct=self.thin_pct)
            loss += self.thin_w * t; logs['thin'] = t.item()

        # ---------- mask BCE ----------
        if self.use_mask_bce and pred_mask is not None:
            inv_mask = 1.0 - mask_f
            bce = F.binary_cross_entropy_with_logits(
                pred_mask,
                inv_mask,
                reduction="mean"
            )
            loss += self.mask_weight * bce; logs['mask_bce'] = bce.item()

        # any optional smoothness or normal terms can be added here
        logs['total'] = loss.item()
        return loss, logs
    
# ──────────────────────────────────────────────────────────────────
class SiLogLoss(nn.Module):
    def __init__(self, lambd: float = .5):
        super().__init__(); self.lambd = lambd
    def forward(self, p: torch.Tensor, t: torch.Tensor, m: torch.Tensor):
        pl, tl = torch.log(p+1e-6), torch.log(t+1e-6)
        d = (pl - tl) * m
        mu = d.mean(); var = (d**2).mean() - self.lambd * mu * mu
        return torch.sqrt(torch.clamp(var, min=0.))



import torch
from torch import nn



# https://hiddenlayers.tech/blog/the-mystery-of-silog-loss/
class _SiLogLoss(nn.Module):
    def __init__(self, lambd=0.5):
        super().__init__()
        self.lambd = lambd

    def forward(self, pred, target, valid_mask):
        valid_mask = valid_mask.detach()
        diff_log = torch.log(target[valid_mask]) - torch.log(pred[valid_mask])
        loss = torch.pow(diff_log, 2).mean() - \
            self.lambd * torch.pow(diff_log.mean(), 2)

        return loss

import logging

logger = logging.getLogger("LossDebug")
logger.setLevel(logging.DEBUG)
handler = logging.StreamHandler()
formatter = logging.Formatter('[%(asctime)s] [%(levelname)s] %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)

import torch
import torch.nn as nn

class SiLogLoss(nn.Module):
    def __init__(self, lambd=0.5):
        super().__init__()
        self.lambd = lambd

    def forward(self, pred: torch.Tensor, target: torch.Tensor, valid_mask: torch.Tensor):
        """
        Computes the SiLog loss without slicing tensors — safe for gradient accumulation.
        
        Args:
            pred (Tensor): (B, 1, H, W) predicted depth
            target (Tensor): (B, 1, H, W) ground truth depth
            valid_mask (Tensor): (B, 1, H, W) boolean or 0/1 mask

        Returns:
            loss (Tensor): scalar
        """

        valid_mask = valid_mask.detach().float()  # Ensure it's float (0s and 1s)

        # Full diff log
        diff_log = torch.log(target + 1e-8) - torch.log(pred + 1e-8)

        # Masked computations
        diff_log = diff_log * valid_mask

        # Normalize by valid pixel count
        n_valid = valid_mask.sum() + 1e-8  # Avoid division by zero

        mean_squared = (diff_log ** 2).sum() / n_valid
        squared_mean = (diff_log.sum() / n_valid) ** 2

        loss = torch.sqrt(mean_squared - self.lambd * squared_mean)

        return loss
    
class _SiLogLoss(nn.Module):
    def __init__(self, lambd=0.5):
        super().__init__()
        self.lambd = lambd

    def forward(self, pred, target, valid_mask):
        valid_mask = valid_mask.detach()
        pred_valid = pred[valid_mask]
        target_valid = target[valid_mask]

        with torch.no_grad():
            num_nan_pred = torch.isnan(pred_valid).sum().item()
            num_nan_target = torch.isnan(target_valid).sum().item()
            num_neg_pred = (pred_valid < 0).sum().item()
            num_neg_target = (target_valid < 0).sum().item()
            num_zero_pred = (pred_valid == 0).sum().item()
            num_zero_target = (target_valid == 0).sum().item()

            logger.debug(f"[SiLogLoss] valid pixels: {pred_valid.numel()}")
            logger.debug(f"[SiLogLoss] NaNs → pred: {num_nan_pred}, target: {num_nan_target}")
            logger.debug(f"[SiLogLoss] Negatives → pred: {num_neg_pred}, target: {num_neg_target}")
            logger.debug(f"[SiLogLoss] Zeros → pred: {num_zero_pred}, target: {num_zero_target}")

        if pred_valid.shape != target_valid.shape:
            print(f"\n[🔥 SiLogLoss CRITICAL] Masked shape mismatch!")
            print(f"  → pred[mask]:   {pred_valid.shape}")
            print(f"  → target[mask]: {target_valid.shape}")
            print(f"  → original pred shape:   {pred.shape}")
            print(f"  → original target shape: {target.shape}")
            print(f"  → mask sum: {valid_mask.sum().item()} (mask shape = {valid_mask.shape})")
            print(f"  → unique mask values: {torch.unique(valid_mask)}")
            raise ValueError("Masked pred and target shapes differ")
        
        # Use pred_valid and target_valid that you already verified!
        try:
            assert pred_valid.shape == target_valid.shape, (
                f"[SiLogLoss] ❌ Shape mismatch after masking: "
                f"pred[mask]={pred_valid.shape}, target[mask]={target_valid.shape}"
            )
            diff_log = torch.log(target_valid) - torch.log(pred_valid)

        except Exception as e:
            import os
            import uuid
            from torchvision.utils import save_image

            sample_id = str(uuid.uuid4())[:8]
            dump_dir = os.path.join("debug_skips", sample_id)
            os.makedirs(dump_dir, exist_ok=True)

            # Save the raw tensors
            torch.save(pred.detach().cpu(), os.path.join(dump_dir, "pred.pt"))
            torch.save(target.detach().cpu(), os.path.join(dump_dir, "target.pt"))
            torch.save(valid_mask.detach().cpu(), os.path.join(dump_dir, "mask.pt"))

            # Optionally save the mask as an image (rescaled to [0, 1])
            try:
                mask_img = valid_mask.float().cpu().squeeze()
                if mask_img.ndim == 2:  # [H, W]
                    save_image(mask_img.unsqueeze(0), os.path.join(dump_dir, "mask.png"))
            except Exception as im_err:
                print(f"[WARN] Failed to save mask image: {im_err}")

            print(f"[SKIP] Sample failed. Dumped to {dump_dir}. Reason: {e}")
            return torch.tensor(0.0, device=pred.device)

        loss = torch.sqrt(torch.pow(diff_log, 2).mean() -
                        self.lambd * torch.pow(diff_log.mean(), 2))
        return loss



#!/usr/bin/env python3
# Copyright (C) 2024-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).
#
# --------------------------------------------------------
# Preprocessing code for the MegaDepth dataset
# dataset at https://www.cs.cornell.edu/projects/megadepth/
# --------------------------------------------------------
import os
import os.path as osp
import collections
from tqdm import tqdm
import numpy as np
os.environ["OPENCV_IO_ENABLE_OPENEXR"] = "1"
import cv2
import h5py

import path_to_root  # noqa
from dust3r.utils.parallel import parallel_threads
from dust3r.datasets.utils import cropping  # noqa


def get_parser():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--megadepth_dir', required=True)
    parser.add_argument('--precomputed_pairs', required=True)
    parser.add_argument('--output_dir', default='data/megadepth_processed')
    return parser


def main(db_root, pairs_path, output_dir):
    os.makedirs(output_dir, exist_ok=True)

    # load all pairs
    data = np.load(pairs_path, allow_pickle=True)
    scenes = data['scenes']
    images = data['images']
    pairs = data['pairs']

    # enumerate all unique images
    todo = collections.defaultdict(set)
    for scene, im1, im2, score in pairs:
        todo[scene].add(im1)
        todo[scene].add(im2)

    # for each scene, load intrinsics and then parallel crops
    for scene, im_idxs in tqdm(todo.items(), desc='Overall'):
        scene, subscene = scenes[scene].split()
        out_dir = osp.join(output_dir, scene, subscene)
        os.makedirs(out_dir, exist_ok=True)

        # load all camera params
        _, pose_w2cam, intrinsics = _load_kpts_and_poses(db_root, scene, subscene, intrinsics=True)

        in_dir = osp.join(db_root, scene, 'dense' + subscene)
        args = [(in_dir, img, intrinsics[img], pose_w2cam[img], out_dir)
                for img in [images[im_id] for im_id in im_idxs]]
        parallel_threads(resize_one_image, args, star_args=True, front_num=0, leave=False, desc=f'{scene}/{subscene}')

    # save pairs
    print('Done! prepared all pairs in', output_dir)


def resize_one_image(root, tag, K_pre_rectif, pose_w2cam, out_dir):
    if osp.isfile(osp.join(out_dir, tag + '.npz')):
        return

    # load image
    img = cv2.cvtColor(cv2.imread(osp.join(root, 'imgs', tag), cv2.IMREAD_COLOR), cv2.COLOR_BGR2RGB)
    H, W = img.shape[:2]

    # load depth
    with h5py.File(osp.join(root, 'depths', osp.splitext(tag)[0] + '.h5'), 'r') as hd5:
        depthmap = np.asarray(hd5['depth'])

    # rectify = undistort the intrinsics
    imsize_pre, K_pre, distortion = K_pre_rectif
    imsize_post = img.shape[1::-1]
    K_post = cv2.getOptimalNewCameraMatrix(K_pre, distortion, imsize_pre, alpha=0,
                                           newImgSize=imsize_post, centerPrincipalPoint=True)[0]

    # downscale
    img_out, depthmap_out, intrinsics_out, R_in2out = _downscale_image(K_post, img, depthmap, resolution_out=(800, 600))

    # write everything
    img_out.save(osp.join(out_dir, tag + '.jpg'), quality=90)
    cv2.imwrite(osp.join(out_dir, tag + '.exr'), depthmap_out)

    camout2world = np.linalg.inv(pose_w2cam)
    camout2world[:3, :3] = camout2world[:3, :3] @ R_in2out.T
    np.savez(osp.join(out_dir, tag + '.npz'), intrinsics=intrinsics_out, cam2world=camout2world)


def _downscale_image(camera_intrinsics, image, depthmap, resolution_out=(512, 384)):
    H, W = image.shape[:2]
    resolution_out = sorted(resolution_out)[::+1 if W < H else -1]

    image, depthmap, intrinsics_out = cropping.rescale_image_depthmap(
        image, depthmap, camera_intrinsics, resolution_out, force=False)
    R_in2out = np.eye(3)

    return image, depthmap, intrinsics_out, R_in2out


def _load_kpts_and_poses(root, scene_id, subscene, z_only=False, intrinsics=False):
    if intrinsics:
        with open(os.path.join(root, scene_id, 'sparse', 'manhattan', subscene, 'cameras.txt'), 'r') as f:
            raw = f.readlines()[3:]  # skip the header

        camera_intrinsics = {}
        for camera in raw:
            camera = camera.split(' ')
            width, height, focal, cx, cy, k0 = [float(elem) for elem in camera[2:]]
            K = np.eye(3)
            K[0, 0] = focal
            K[1, 1] = focal
            K[0, 2] = cx
            K[1, 2] = cy
            camera_intrinsics[int(camera[0])] = ((int(width), int(height)), K, (k0, 0, 0, 0))

    with open(os.path.join(root, scene_id, 'sparse', 'manhattan', subscene, 'images.txt'), 'r') as f:
        raw = f.read().splitlines()[4:]  # skip the header

    extract_pose = colmap_raw_pose_to_principal_axis if z_only else colmap_raw_pose_to_RT

    poses = {}
    points3D_idxs = {}
    camera = []

    for image, points in zip(raw[:: 2], raw[1:: 2]):
        image = image.split(' ')
        points = points.split(' ')

        image_id = image[-1]
        camera.append(int(image[-2]))

        # find the principal axis
        raw_pose = [float(elem) for elem in image[1: -2]]
        poses[image_id] = extract_pose(raw_pose)

        current_points3D_idxs = {int(i) for i in points[2:: 3] if i != '-1'}
        assert -1 not in current_points3D_idxs, bb()
        points3D_idxs[image_id] = current_points3D_idxs

    if intrinsics:
        image_intrinsics = {im_id: camera_intrinsics[cam] for im_id, cam in zip(poses, camera)}
        return points3D_idxs, poses, image_intrinsics
    else:
        return points3D_idxs, poses


def colmap_raw_pose_to_principal_axis(image_pose):
    qvec = image_pose[: 4]
    qvec = qvec / np.linalg.norm(qvec)
    w, x, y, z = qvec
    z_axis = np.float32([
        2 * x * z - 2 * y * w,
        2 * y * z + 2 * x * w,
        1 - 2 * x * x - 2 * y * y
    ])
    return z_axis


def colmap_raw_pose_to_RT(image_pose):
    qvec = image_pose[: 4]
    qvec = qvec / np.linalg.norm(qvec)
    w, x, y, z = qvec
    R = np.array([
        [
            1 - 2 * y * y - 2 * z * z,
            2 * x * y - 2 * z * w,
            2 * x * z + 2 * y * w
        ],
        [
            2 * x * y + 2 * z * w,
            1 - 2 * x * x - 2 * z * z,
            2 * y * z - 2 * x * w
        ],
        [
            2 * x * z - 2 * y * w,
            2 * y * z + 2 * x * w,
            1 - 2 * x * x - 2 * y * y
        ]
    ])
    # principal_axis.append(R[2, :])
    t = image_pose[4: 7]
    # World-to-Camera pose
    current_pose = np.eye(4)
    current_pose[: 3, : 3] = R
    current_pose[: 3, 3] = t
    return current_pose


if __name__ == '__main__':
    parser = get_parser()
    args = parser.parse_args()
    main(args.megadepth_dir, args.precomputed_pairs, args.output_dir)



import os
import json
import torch
import torchvision.transforms as transforms
import os.path
import numpy as np
import cv2
from torch.utils.data import Dataset
import random
from .__base_dataset__ import BaseDataset


class FisheyeDataset(BaseDataset):
    def __init__(self, cfg, phase, **kwargs):
        super(FisheyeDataset, self).__init__(
            cfg=cfg,
            phase=phase,
            **kwargs)
        self.metric_scale = cfg.metric_scale
    
    def load_data(self, path: str, is_rgb_img: bool=False):
        if not os.path.exists(path):
            self.logger.info(f'>>>>{path} does not exist.')
            # raise RuntimeError(f'{path} does not exist.')

        data_type = os.path.splitext(path)[-1]
        if data_type in self.img_file_type:
            if is_rgb_img:
                data = cv2.imread(path)
            else:
                data = cv2.imread(path, -1)
                data[data>65500] = 0
                data &= 0x7FFF

        elif data_type in self.np_file_type:
            data = np.load(path)
        else:
            raise RuntimeError(f'{data_type} is not supported in current version.')
        
        return data.squeeze()

    def load_batch(self, meta_data, data_path):
        curr_intrinsic = meta_data['cam_in']
        # load rgb/depth
        curr_rgb, curr_depth = self.load_rgb_depth(data_path['rgb_path'], data_path['depth_path'])
        # get semantic labels
        curr_sem = self.load_sem_label(data_path['sem_path'], curr_depth)
        # create camera model
        curr_cam_model = self.create_cam_model(curr_rgb.shape[0], curr_rgb.shape[1], curr_intrinsic)       
        # get normal labels
        curr_normal = self.load_norm_label(data_path['normal_path'], H=curr_rgb.shape[0], W=curr_rgb.shape[1]) 
        # get depth mask
        depth_mask = self.load_depth_valid_mask(data_path['depth_mask_path'])[:, :, :]
         
        # with masks from andy
        curr_depth[~(depth_mask[:, :, 0])] = -1
        curr_rgb[~(depth_mask[:, :, :])] = 0
        
        # get stereo depth
        curr_stereo_depth = self.load_stereo_depth_label(data_path['disp_path'], H=curr_rgb.shape[0], W=curr_rgb.shape[1]) 

        data_batch = dict(
            curr_rgb = curr_rgb,
            curr_depth = curr_depth,
            curr_sem = curr_sem,
            curr_normal = curr_normal,
            curr_cam_model=curr_cam_model,
            curr_stereo_depth=curr_stereo_depth,
        )
        return data_batch

    
    def process_depth(self, depth, rgb):

        depth /= self.metric_scale
        return depth



import torch
import torch.nn as nn

class ScaleAlignLoss(nn.Module):
    """
    Loss function defined over sequence of depth predictions
    """
    def __init__(self, data_type=['lidar', 'denselidar', 'stereo', 'denselidar_syn'], loss_weight=1.0, disable_dataset=['MapillaryPSD'], **kwargs):
        super(ScaleAlignLoss, self).__init__()
        self.loss_weight = loss_weight
        self.data_type = data_type
        self.disable_dataset = disable_dataset

    def forward(self, prediction, target, mask, scale, **kwargs):
        device = target.device

        B, C, H, W = prediction.shape


        # median_pred, _ = torch.median(prediction.view(B, C*H*W), 1)
        # median_pred = median_pred.detach()

        # scale_factor = torch.zeros_like(scale).squeeze(3).squeeze(2).squeeze(1)
        # for i in range(B):
        #     mask_i = mask[i, ...]
        #     if torch.sum(mask_i) > 10:
        #         scale_factor[i] = torch.median(target[i, ...][mask_i]) / (torch.median(prediction[i, ...][mask_i]) + 1e-8)
        #     else:
        #         scale_factor[i] = 0
        
        # target_scale = (median_pred * scale_factor)

        # batches_dataset = kwargs['dataset']
        # self.batch_valid = torch.tensor([1 if batch_dataset not in self.disable_dataset else 0 \
        #     for batch_dataset in batches_dataset], device=device)

        # batch_valid = self.batch_valid * (scale_factor > 1e-8)

        # scale_diff = torch.abs(scale.squeeze(3).squeeze(2).squeeze(1) - scale_factor * median_pred)

        batches_dataset = kwargs['dataset']
        self.batch_valid = torch.tensor([1 if batch_dataset not in self.disable_dataset else 0 \
            for batch_dataset in batches_dataset], device=device)

        scale_tgt = torch.zeros_like(scale).squeeze(3).squeeze(2).squeeze(1)
        for i in range(B):
            mask_i = mask[i, ...]
            if torch.sum(mask_i) > 10:
                scale_tgt[i] = torch.median(target[i, ...][mask_i])
            else:
                scale_tgt[i] = 0
        
        batch_valid = self.batch_valid * (scale_tgt > 1e-8)
        scale_diff = torch.abs(scale.squeeze(3).squeeze(2).squeeze(1) - scale_tgt)
        loss = torch.sum(scale_diff * batch_valid) / (torch.sum(batch_valid) + 1e-8)

        return loss * self.loss_weight


# MIT License

# Copyright (c) 2022 Intelligent Systems Lab Org

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# File author: Shariq Farooq Bhat

import os

import numpy as np
import torch
from PIL import Image
from torch.utils.data import DataLoader, Dataset
from torchvision import transforms


class ToTensor(object):
    def __init__(self, resize_shape):
        # self.normalize = transforms.Normalize(
        #     mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        self.normalize = lambda x : x
        self.resize = transforms.Resize(resize_shape)

    def __call__(self, sample):
        image, depth = sample['image'], sample['depth']
        image = self.to_tensor(image)
        image = self.normalize(image)
        depth = self.to_tensor(depth)

        image = self.resize(image)

        return {'image': image, 'depth': depth, 'dataset': "ddad"}

    def to_tensor(self, pic):

        if isinstance(pic, np.ndarray):
            img = torch.from_numpy(pic.transpose((2, 0, 1)))
            return img

        #         # handle PIL Image
        if pic.mode == 'I':
            img = torch.from_numpy(np.array(pic, np.int32, copy=False))
        elif pic.mode == 'I;16':
            img = torch.from_numpy(np.array(pic, np.int16, copy=False))
        else:
            img = torch.ByteTensor(
                torch.ByteStorage.from_buffer(pic.tobytes()))
        # PIL image mode: 1, L, P, I, F, RGB, YCbCr, RGBA, CMYK
        if pic.mode == 'YCbCr':
            nchannel = 3
        elif pic.mode == 'I;16':
            nchannel = 1
        else:
            nchannel = len(pic.mode)
        img = img.view(pic.size[1], pic.size[0], nchannel)

        img = img.transpose(0, 1).transpose(0, 2).contiguous()

        if isinstance(img, torch.ByteTensor):
            return img.float()
        else:
            return img


class DDAD(Dataset):
    def __init__(self, data_dir_root, resize_shape):
        import glob

        # image paths are of the form <data_dir_root>/{outleft, depthmap}/*.png
        self.image_files = glob.glob(os.path.join(data_dir_root, '*.png'))
        self.depth_files = [r.replace("_rgb.png", "_depth.npy")
                            for r in self.image_files]
        self.transform = ToTensor(resize_shape)

    def __getitem__(self, idx):

        image_path = self.image_files[idx]
        depth_path = self.depth_files[idx]

        image = np.asarray(Image.open(image_path), dtype=np.float32) / 255.0
        depth = np.load(depth_path)  # meters

        # depth[depth > 8] = -1
        depth = depth[..., None]

        sample = dict(image=image, depth=depth)
        sample = self.transform(sample)

        if idx == 0:
            print(sample["image"].shape)

        return sample

    def __len__(self):
        return len(self.image_files)


def get_ddad_loader(data_dir_root, resize_shape, batch_size=1, **kwargs):
    dataset = DDAD(data_dir_root, resize_shape)
    return DataLoader(dataset, batch_size, **kwargs)



# MIT License

# Copyright (c) 2022 Intelligent Systems Lab Org

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# File author: Shariq Farooq Bhat

import gradio as gr
from zoedepth.utils.misc import colorize
from PIL import Image
import tempfile

def predict_depth(model, image):
    depth = model.infer_pil(image)
    return depth

def create_demo(model):
    gr.Markdown("### Depth Prediction demo")
    with gr.Row():
        input_image = gr.Image(label="Input Image", type='pil', elem_id='img-display-input').style(height="auto")
        depth_image = gr.Image(label="Depth Map", elem_id='img-display-output')
    raw_file = gr.File(label="16-bit raw depth, multiplier:256")
    submit = gr.Button("Submit")

    def on_submit(image):
        depth = predict_depth(model, image)
        colored_depth = colorize(depth, cmap='gray_r')
        tmp = tempfile.NamedTemporaryFile(suffix='.png', delete=False)
        raw_depth = Image.fromarray((depth*256).astype('uint16'))
        raw_depth.save(tmp.name)
        return [colored_depth, tmp.name]
    
    submit.click(on_submit, inputs=[input_image], outputs=[depth_image, raw_file])
    # examples = gr.Examples(examples=["examples/person_1.jpeg", "examples/person_2.jpeg", "examples/person-leaves.png", "examples/living-room.jpeg"],
    #                        inputs=[input_image])


from _typeshed import Incomplete

logger: Incomplete

class RegistryBase(type):
    REGISTRY: dict[str, type]
    def __new__(cls, name, bases, attrs): ...
    @classmethod
    def get_registry(cls) -> dict[str, type]: ...

class BaseRegisteredClass(metaclass=RegistryBase): ...
